-- don't load if class is wrong
local _, class = UnitClass("player")
if class ~= "SHAMAN" then return end
local GetTime = GetTime
local debug = clcInfo.debug

local version = 50400

local modName = "__enhancement"
local mod = clcInfo:RegisterClassModule(modName)
local emod = clcInfo.env
local db -- template based

local ef = CreateFrame("Frame") 
ef:Hide()

local defaults = {
	version = version,
	prio = "ls wfw ftw st0 eb ufue lifeblood mw5lb sw ss fsnofs ll mw4lbt142p fs ue aslb es st fn",
	prioAoe = "ls wfw ftw st0 ufue eb mw5lb sw ss fsnofs ll mw4lbt142p fs ue aslb es st fn"
}

-- @defines
--------------------------------------------------------------------------------
local gcdId 				= 403 -- lb for gcd
-- list of spellId
local wfwId					= 8232 -- windfury weapon
local swId					= 51533 -- feral spirits
local llId					= 60103 -- laval lash
local ssId					= 17364 -- stormstrike
local fnId 					= 1535	-- fire nova
local ebId					= 117014 -- elemental blast
local hrId					= 73920 -- Healing rain
local hstId					= 5394 -- Healing Stream Totem

local lvbId 				= 51505	-- lava burst
local lsId 					= 324	-- lightning shield
local fsId					= 8050	-- flame shock
local esId					= 8042	-- earth shock
local lbId					= 403	-- lightning bolt
local stId					= 3599	-- searing totem
local ftwId					= 8024	-- flametongue weapon
local clId					= 421 	-- chain lightning
local asId					= 16188 -- ancestral swiftness
local swgId					= 79206 -- spiritwalker's grace
local ueId					= 73680 -- unleash elements
local lifebloodId			= 121279 -- lifeblood
-- fsName to pass as argument for range range checks
local fsName				= GetSpellInfo(fsId)
-- buffs
local buffLS 				= GetSpellInfo(lsId)	-- lightning shield
local buffUE				= GetSpellInfo(ueId)		-- unleash elements
local buffTW				= GetSpellInfo(77746)		-- totemic wrath
local buffFTW 				= GetSpellInfo(ftwId)		-- flametongue weapon
local buffAS				= GetSpellInfo(asId)	-- Ancestral swiftness
local debuffFS				= GetSpellInfo(fsId)     -- Flame Shock


local buffUEF				= GetSpellInfo(73683) -- flametongue unleash
local buffMW				= GetSpellInfo(53817) -- maelstrom weapon

-- status vars
local s1, s2, s3
local s_ctime, s_otime, s_gcd, s_lsc, s_ftw, s_wfw, s_as, s_lsd, s_fsd, ss_name, s_fscd, s_uef, s_mws, s_std, s_mana, s_haste, s_castingid, s_castingdur, fe_active

function mod:OnTemplatesUpdate()
	db = clcInfo:RegisterClassModuleTDB(modName, defaults)
	if db then
		if not db.version or db.version < version then
			-- fix stuff
			clcInfo.AdaptConfigAndClean(modName, db, defaults)
			db.version = version
		end
		
		mod:UpdateQueue()
	end
end

-- the queue
local qn = {} 		-- normal queue
local qa = {} 		-- aoe queue
local q 					-- working var

local function GetCooldown(id)
	local start, duration = GetSpellCooldown(id)
	local cd = duration - (s_ctime - start) - max(s_castingdur, s_gcd)
	cd = math.floor(cd * 10000) / 10000
	if cd <= 0.0 then return 0.0 end
	return cd
end

function GetCooldownDuration(id)
	if id == lvbId then
		return 8.0
	elseif id == eqId then
		return 10.0
	else
		return 1.5 / s_haste
	end
end

function GetCastTime(id)
	local castTime
	_, _, _, _, _, _, castTime = GetSpellInfo(id)
	castTime = castTime / 1000
	return castTime 
end

function GetTalentAvailability(id)
	local name, _, _, _, selected, available = GetTalentInfo(id)
	return (name and selected and available)
end

function t16Check()
	slots = {"Grips", "Helmet", "Spaulders", "Cuirass", "Legguards"}
	tier = 0
	for i=1,5,1 do
		if(IsEquippedItem(slots[i] .. " of Celestial Harmony")) then
			tier = tier + 1
		end
	end
	return tier
end
	

-- actions ---------------------------------------------------------------------
local actions = {
	
	ls = {
		id = lsId,
		GetCD = function()
			if s_lsd <= 0 then
				return 0.0
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
			s_lsd = 100
			s_lsc = 3
		end,
		info = "apply Lightning Shield",
	},
	ftw = {
		id = ftwId,
		GetCD = function()
			if s_ftw <= 0 then
				return 0.0
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
			s_ftw = 100	-- make sure it's not shown for next skill
		end,
		info = "Apply Flametongue Weapon",
	},
	wfw = {
		id = wfwId,
		GetCD = function()
			if s_wfw <= 0 then
				return 0.0
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
			s_wfw = 100	-- make sure it's not shown for next skill
		end,
		info = "Apply Windfury Weapon",
	},
	st = {
		id = stId,
		GetCD = function()
			if s_std <= 5 and not fe_active then
				return 0.0
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.0
			s_std = 60
		end,
		info = "use Searing Totem if 5 or less seconds left",
	},
	st0 = {
		id = stId,
		GetCD = function()
			if s_std <= 0 then
				return 0.0
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.0
			s_std = 60
		end,
		info = "use Searing Totem if 0 seconds left",
	},
	fsnofs = {
		id = fsId,
		GetCD = function()
			if s1 ~= fsId and s_uef > 0 and s_fsd == 0 then 
				return GetCooldown(fsId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
			s_fsd = 100
		end,
		info = "Flame Shock",
	},
	fs = {
		id = fsId,
		GetCD = function()
			if s1 ~= fsId and (s_fsd == 0 or (s_uef > 0 and (s_fsd < 10 or t16Check() >= 2))) then 
				return GetCooldown(fsId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
			s_fsd = 100
		end,
		info = "FS refresh",
	},
	lvb = {
		id = lvbId,
		GetCD = function()
			if s_fsd <= 0 then
				return 100
			elseif s_castingid == lvbId then
				return GetCooldownDuration(lvbId) + s_castingdur 
			elseif s1 ~= lvbId then
				return GetCooldown(lvbId)
			end
			return 99
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(lvbId) / s_haste
		end,
		info = "Lava Burst",
	},
	fn = {
		id = fnId,
		GetCD = function()
			local fnc = GetCooldown(fnId)
			if s1 ~= fnId then
				if fnc <= 0.5 then
					return 0
				else
					return GetCooldown(fnId)
				end
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
		end,
		info = "Fire Nova",
	},
	es = {
		id = esId,
		GetCD = function()
			if s1 ~= esId and s1 ~= fsId then
				return GetCooldown(esId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
		end,
		info = "Earth Shock",
	},
	lb = {
		id = lbId,
		GetCD = function()
			return 0.0
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(lbId) / s_haste
		end,
		info = "Lightning Bolt",
	},
	eb = {
		id = ebId,
		GetCD = function()
			if s1 ~= ebId and s_mws >= 1 and GetTalentAvailability(18) then
				return GetCooldown(ebId)
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(ebId) / s_haste
			s_mws = 0
		end,
		info = "Elemental Blast",
	},
	mw5lb = {
		id = lbId,
		GetCD = function()
			if s1 ~= lbId and s_mws >= 5 then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(lbId) / s_haste
			s_mws = 0
		end,
		info = "MW5 Lightning Bolt",
	},
	aslb = {
		id = lbId,
		GetCD = function()
			if s1 ~= lbId and GetTalentAvailability(11) then
				if s_as > 0 or GetCooldown(asId) == 0 then 
					return 0 
				else
					return GetCooldown(asId)
				end
			else
				return 105
			end
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(lbId) / s_haste
			s_as = 0
		end,
		info = "AS Lightning Bolt",
	},
	mw4lbt142p = {
		id = lbId,
		GetCD = function()
			if s1 ~= lbId and s_mws >= 4 and ss_name ~= "Stormblast" then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(lbId) / s_haste
			s_mws = 0
		end,
		info = "MW4 Lightning Bolt if T14 2pc",
	},
	mw4lb = {
		id = lbId,
		GetCD = function()
			if s1 ~= lbId and s_mws >= 4 and ss_name ~= "Stormblast" then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(lbId) / s_haste
			s_mws = 0
		end,
		info = "MW4 Lightning Bolt",
	},
	mw3lb = {
		id = lbId,
		GetCD = function()
			if s1 ~= lbId and s_mws >= 3 and ss_name ~= "Stormblast" then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(lbId) / s_haste
			s_mws = 0
		end,
		info = "MW3 Lightning Bolt",
	},
	mw2lb = {
		id = lbId,
		GetCD = function()
			if s1 ~= lbId and s_mws >= 2 and ss_name ~= "Stormblast" then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(lbId) / s_haste
			s_mws = 0
		end,
		info = "MW2 Lightning Bolt",
	},
	mw1lb = {
		id = lbId,
		GetCD = function()
			if s1 ~= lbId and s_mws >= 5 and ss_name ~= "Stormblast" then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(lbId) / s_haste
			s_mws = 0
		end,
		info = "MW1 Lightning Bolt",
	},
	mw5cl = {
		id = clId,
		GetCD = function()
			if s1 ~= clId and s_mws >= 5 then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(clId) / s_haste
			s_mws = 0
		end,
		info = "MW5 Chain Lightning",
	},
	mw4cl = {
		id = clId,
		GetCD = function()
			if s1 ~= clId and s_mws >= 4 and ss_name ~= "Stormblast" then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(clId) / s_haste
			s_mws = 0
		end,
		info = "MW4 Chain Lightning",
	},
	mw3cl = {
		id = clId,
		GetCD = function()
			if s1 ~= clId and s_mws >= 3 and ss_name ~= "Stormblast" then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(clId) / s_haste
			s_mws = 0
		end,
		info = "MW3 Chain Lightning",
	},
	mw2cl = {
		id = clId,
		GetCD = function()
			if s1 ~= clId and s_mws >= 2 and ss_name ~= "Stormblast" then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(clId) / s_haste
			s_mws = 0
		end,
		info = "MW2 Chain Lightning",
	},
	mw1cl = {
		id = clId,
		GetCD = function()
			if s1 ~= clId and s_mws >= 1 and ss_name ~= "Stormblast" then
				return 0
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(clId) / s_haste
			s_mws = 0
		end,
		info = "MW1 Chain Lightning",
	},
	cl = {
		id = clId,
		GetCD = function()
			return 0.0
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(clId) / s_haste
		end,
		info = "Chain Lightning",
	},	
	ss = {
		id = ssId,
		GetCD = function()
			if s1 ~= ssId then
				return GetCooldown(ssId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5
		end,
		info = "Stormstrike",
	},
	ll = {
		id = llId,
		GetCD = function()
			if s1 ~= llId then
				return GetCooldown(llId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5
		end,
		info = "Lava Lash",
	},
	sw = {
		id = swId,
		GetCD = function()
			if s1 ~= swId then
				return GetCooldown(swId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
		end,
		info = "Spirit Wolves",
	},
	ue = {
		id = ueId,
		GetCD = function()
			if s1 ~= ueId and not GetTalentAvailability(16) then
				return GetCooldown(ueId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
			s_uef = 8
		end,
		info = "Unleash Elements",
	},
	ufue = {
		id = ueId,
		GetCD = function()
			if s1 ~= ueId and GetTalentAvailability(16) then
				return GetCooldown(ueId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
			s_uef = 8
		end,
		info = "Unleash Elements",
	},
	ufT162pc = {
		id = ueId,
		GetCD = function()
			if s1 ~= ueId and t16Check() > 2 then
				return GetCooldown(ueId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
			s_uef = 8
		end,
		info = "Unleash Elements",
	},
	lifeblood = {
		id = lifebloodId,
		GetCD = function()
			if s1 ~= lifebloodId and GetCooldown("Fire Elemental Totem") > 70 then
				return GetCooldown(lifebloodId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime
		end,
		info = "Lifeblood",
	},
	hrmw5 = {
		id = hrId,
		GetCD = function()
			if s1 ~= hrId and s_mws >= 5 then
				return GetCooldown("Healing Rain")
			end
			return 101
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + GetCastTime(hrId) / s_haste
			s_mws = 0
		end,
		info = "MW5 Healing Rain",
	},
	hst = {
		id = hstId,
		GetCD = function()
			if s1 ~= hstId then
				return GetCooldown(hstId)
			end
			return 100
		end,
		UpdateStatus = function()
			s_ctime = s_ctime + 1.5 / s_haste
		end,
		info = "healing stream",
	},	
}
mod.actions = actions
--------------------------------------------------------------------------------

function mod:UpdateQueue()
	-- normal queue
	qn = {}
	for v in string.gmatch(db.prio, "[^ ]+") do
		if actions[v] then
			table.insert(qn, v)
		else
			print("clcInfo", modName, "invalid action:", v)
		end
	end
	db.prio = table.concat(qn, " ")
	
	-- aoe queue
	qa = {}
	for v in string.gmatch(db.prioAoe, "[^ ]+") do
		if actions[v] then
			table.insert(qa, v)
		else
			print("clcInfo", modName, "invalid action (aoe prio):", v)
		end
	end
	db.prioAoe = table.concat(qa, " ")
end

local function GetBuff(buff)
	local left
	_, _, _, _, _, _, expires = UnitBuff("player", buff, nil, "PLAYER")
	if expires then
		left = expires - s_ctime - max(s_castingdur, s_gcd)
		if left < 0 then left = 0 end
	else
		left = 0
	end
	return left
end

local function GetDebuff(debuff)
	local left
	_, _, _, _, _, _, expires, caster, _, _, _ = UnitDebuff("target", debuff)
	if caster ~= "player" then
		left = 0
	elseif expires then
		left = expires - s_ctime - max(s_castingdur, s_gcd)
		if left < 0 then left = 0 end
	else
		left = 0
	end
	return left
end

-- reads all the interesting data
local function GetStatus()
	-- current time
	s_ctime = GetTime()

	-- currently being cast spell
	local stop, name
	s_castingdur = 0
	name, _, _, _, _, stop, _, _, _ = UnitCastingInfo("player")
	if name ~= nil then
		_, s_castingid = GetSpellBookItemInfo(name)
	else
		s_castingid = 0
	end
	
	if stop ~= nil then
		s_castingdur = (stop - s_ctime * 1000) / 1000
		if s_castingdur <= 0 then
			s_castingdur = 0
		end
	end
	
	-- gcd value
	local start, duration = GetSpellCooldown(gcdId)
	s_gcd = start + duration - s_ctime
	if s_gcd < 0 then s_gcd = 0 end
	
	-- the buffs
	_, s_wfw, _, _, s_ftw, _ = GetWeaponEnchantInfo()
	if s_ftw == nil then
		s_ftw = 0
	end
	if s_wfw == nil then
		s_wfw = 0
	end
	s_lsd = GetBuff(buffLS)
	s_uef = GetBuff(buffUEF)

	-- client lightning shield charges
	_, _, _, s_lsc, _, _, _, _, _, _, _ = UnitBuff("player", "Lightning Shield")
	if s_lsc == nil then
		s_lsc = 0
	end
	
	-- client mw charges
	_, _, _, s_mws, _, _, _, _, _, _, _ = UnitBuff("player", "Maelstrom Weapon")
	if s_mws == nil then
		s_mws = 0
	end
	_, _, _, s_as, _, _, _, _, _, _, _ = UnitBuff("player", "Ancestral Swiftness")
	if s_as == nil then
		s_as = 0
	end	
	_,fire_totem,_,_ = GetTotemInfo(1)
	fe_active = fire_totem == "Fire Elemental Totem"
	for i = 73, 77 do
		ss_name,_ = GetSpellBookItemName( i, BOOKTYPE_SPELL );
		if ss_name == "Stormstrike" or ss_name == "Stormblast" then
			break;
		end
	end
	--print(ss_name)
	-- haste
	s_haste = 1 + UnitSpellHaste("player") / 100
	
	-- flame shock duration
	s_fsd = GetDebuff(debuffFS)

	-- flame shock cd
	s_fscd = GetCooldown(fsId)
	
	-- searing totem duration
	s_std = GetTotemTimeLeft(1)
	
	-- mana level
	s_mana = UnitPower("player")
end

local function GetNextAction(queue)
	-- decide which queue to use
	q = queue
	-- TODO add AOE queue

	local n = #q
	
	-- parse once, get cooldowns, return first 0
	for i = 1, n do
		local action = actions[q[i]]
		local cd = action.GetCD()
		if debug.enabled then
			debug:AddBoth(q[i], cd)
		end
		if cd == 0 then
			return action.id, q[i]
		end
		action.cd = cd
	end
	
	-- parse again, return min cooldown
	local minQ = 1
	local minCd = actions[q[1]].cd
	for i = 2, n do
		local action = actions[q[i]]
		if minCd > action.cd then
			minCd = action.cd
			minQ = i
		end
	end
	return actions[q[minQ]].id, q[minQ]
end

local function EnhRotation()
	s1 = nil
	s3 = nil
	GetStatus()
	if debug.enabled then
		debug:Clear()
		--[[debug:AddBoth("ctime", s_ctime)
		debug:AddBoth("gcd", s_gcd)
		debug:AddBoth("lsc", s_lsc)
		debug:AddBoth("fsd", s_fsd)
		debug:AddBoth("fscd", s_fscd)
		debug:AddBoth("ftw", s_ftw)
		debug:AddBoth("lsd", s_lsd)
		debug:AddBoth("std", s_std)
		debug:AddBoth("haste", s_haste)
		debug:AddBoth("castingid", s_castingid)
		debug:AddBoth("castingdur", s_castingdur)
		debug:AddBoth("lvbCDdur", GetCooldownDuration(lvbId))
		]]--
	end
	local action
	s1, action = GetNextAction(qn)
	if debug.enabled then
		debug:AddBoth("s1", action)
	end
	-- 
	s_otime = s_ctime -- save it so we adjust buffs for next
	actions[action].UpdateStatus()

	-- adjust buffs
	s_otime = s_ctime - s_otime
	s_fsd = max(0, s_fsd - s_otime)
	s_fscd = max(0, s_fscd - s_otime)
	s_ftw = max(0, s_ftw - s_otime)
	s_lsd = max(0, s_lsd - s_otime)
	s_std = max(0, s_std - s_otime)

	if debug.enabled then
		--[[debug:AddBoth("ctime", s_ctime)
		debug:AddBoth("otime", s_otime)
		debug:AddBoth("gcd", s_gcd)
		debug:AddBoth("lsc", s_lsc)
		debug:AddBoth("fsd", s_fsd)
		debug:AddBoth("fscd", s_fscd)
		debug:AddBoth("ftw", s_ftw)
		debug:AddBoth("lsd", s_lsd)
		debug:AddBoth("std", s_std)
		debug:AddBoth("haste", s_haste)
		debug:AddBoth("castingid", s_castingid)
		debug:AddBoth("castingdur", s_castingdur)
		debug:AddBoth("lvbCDdur", GetCooldownDuration(lvbId))
		]]--
	end
	s_s2 = true
	s2, action = GetNextAction(qn)
	if debug.enabled then
		debug:AddBoth("s2", action)
	end
	GetStatus()
	s3, action = GetNextAction(qa)
	if debug.enabled then
		debug:AddBoth("s3", action)
	end
end

-- plug in
--------------------------------------------------------------------------------
local secondarySkill
local aoeSkill
function emod.IconEnh1(...)
	EnhRotation(...)
	if secondarySkill then secondarySkill:DoUpdate() end
	if aoeSkill then aoeSkill:DoUpdate() end
	return emod.IconSpell(s1, db.rangePerSkill or fsName)
end

local function SecondaryExec()
	return emod.IconSpell(s2, db.rangePerSkill or fsName)
end
local function ExecCleanup2()
	secondarySkill = nil
end
function emod.IconEnh2(...)
	secondarySkill = emod.___e
	secondarySkill:SetScript("OnUpdate", nil)
	secondarySkill.exec = SecondaryExec
	secondarySkill.ExecCleanup = ExecCleanup2
end

local function AoeExec()
	return emod.IconSpell(s3, db.rangePerSkill or fsName)
end
local function ExecCleanupAoe()
	aoeSkill = nil
end
function emod.IconEnhAoe(...)
	aoeSkill = emod.___e
	aoeSkill:SetScript("OnUpdate", nil)
	aoeSkill.exec = AoeExec
	aoeSkill.ExecCleanup = ExecCleanupAoe
end
-- giving queue as command line
local function CmdEnh(args)
	db.prio = table.concat(args, " ")
	mod:UpdateQueue()
	clcInfo:UpdateOptions()
end
clcInfo.cmdList["enhprio"] = CmdEnh

-- for zeal too
local function CmdEnhAoe(args)
	db.prioAoe = table.concat(args, " ")
	mod:UpdateQueue()
	clcInfo:UpdateOptions()
end
clcInfo.cmdList["enhprioaoe"] = CmdEnhAoe